﻿Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Linq
Imports System.Text

Namespace SolSoft.DataBinding
    ''' <summary>
    ''' Represents a property whose value is derived from one or more other properties.
    ''' </summary>
    ''' <typeparam name="TValue">Type of the property's value</typeparam>
    Public Class DerivedNotifyProperty(Of TValue)
        Implements IProperty(Of TValue)
        Private ReadOnly m_owner As IRaisePropertyChanged
        Private ReadOnly m_derivedPropertyName As String
        Private ReadOnly m_getValueProperty As Func(Of TValue)
        Private ReadOnly m_dependentValuesToListenFor As Dictionary(Of String, Boolean)

        Private Sub New(ByVal owner As IRaisePropertyChanged, ByVal derivedPropertyName As String, ByVal getDerivedPropertyValue As Func(Of TValue))
            If owner Is Nothing Then
                Throw New ArgumentNullException(ObjectNamingExtensions.GetName(Function() owner))
            End If
            Me.m_owner = owner

            If String.IsNullOrEmpty(derivedPropertyName) Then
                Throw New ArgumentException(ObjectNamingExtensions.GetName(Function() derivedPropertyName))
            End If
            Me.m_derivedPropertyName = derivedPropertyName

            If getDerivedPropertyValue Is Nothing Then
                Throw New ArgumentNullException(ObjectNamingExtensions.GetName(Function() getDerivedPropertyValue))
            End If
            Me.m_getValueProperty = getDerivedPropertyValue
            Me.m_value = New Lazy(Of TValue)(m_getValueProperty)
        End Sub

        ''' <summary>
        ''' Creates a new <see cref="DerivedNotifyProperty&lt;TValue&gt;"/>.  When any of the specified properties change in the owner object, this property's value will be re-calculated the next time it is requested
        ''' </summary>
        ''' <param name="owner">Object that contains this property and the ones from which it is derived</param>
        ''' <param name="derivedPropertyName">Name of this property in the owner</param>
        ''' <param name="getDerivedPropertyValue">Function to determine the current value of this property</param>
        ''' <param name="propertiesDependedOn">Properties that this value depends on.  When any of these properties change, this property's value will be re-calculated the next time it is requested.  
        ''' These properties must therefore be in the same owner object as this one.</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal derivedPropertyName As String, ByVal getDerivedPropertyValue As Func(Of TValue), ParamArray ByVal propertiesDependedOn() As String)
            Me.New(owner, derivedPropertyName, getDerivedPropertyValue)
            AddHandler Me.m_owner.PropertyChanged, AddressOf m_owner_PropertyChanged

            If propertiesDependedOn Is Nothing Then
                Throw New ArgumentNullException(ObjectNamingExtensions.GetName(Function() propertiesDependedOn))
            End If
            Me.m_dependentValuesToListenFor = propertiesDependedOn.ToDictionary(Function([property]) [property], Function([property]) True)
        End Sub

        ''' <summary>
        ''' Creates a new <see cref="DerivedNotifyProperty&lt;TValue&gt;"/>.  When any of the specified properties change, this property's value will be re-calculated the next time it is requested
        ''' </summary>
        ''' <param name="owner">Object that contains this property</param>
        ''' <param name="derivedPropertyName">Name of this property in the owner</param>
        ''' <param name="getDerivedPropertyValue">Function to determine the current value of this property</param>
        ''' <param name="valueChangesToListenFor">Properties that this value depends on.  When any of these properties change, this property's value will be re-calculated the next time it is requested.</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal derivedPropertyName As String, ByVal getDerivedPropertyValue As Func(Of TValue), ParamArray ByVal valueChangesToListenFor() As INotifyValueChanged)
            Me.New(owner, derivedPropertyName, getDerivedPropertyValue)
            If valueChangesToListenFor Is Nothing Then
                Throw New ArgumentNullException(ObjectNamingExtensions.GetName(Function() valueChangesToListenFor))
            End If

            For Each valueChangeToListenFor As INotifyValueChanged In valueChangesToListenFor
                AddHandler valueChangeToListenFor.ValueChanged, AddressOf TriggerValueChanged
            Next valueChangeToListenFor
        End Sub

        ''' <summary>
        ''' Gets the name of this property
        ''' </summary>
        Public ReadOnly Property Name() As String Implements IProperty(Of TValue).Name
            Get
                Return m_derivedPropertyName
            End Get
        End Property

        Private m_value As Lazy(Of TValue)

        ''' <summary>
        ''' Gets the current value of this property
        ''' </summary>
        Public ReadOnly Property Value() As TValue Implements IProperty(Of TValue).Value
            Get
                Return m_value.Value
            End Get
        End Property

        ''' <summary>
        ''' Event fired when the property value has changed
        ''' </summary>
        Public Event ValueChanged As EventHandler(Of ValueChangedEventArgs) Implements SolSoft.DataBinding.INotifyValueChanged.ValueChanged

        ''' <summary>
        ''' Raises this object's <see cref="ValueChanged"/> event
        ''' </summary>
        ''' <param name="e"></param>
        Protected Sub OnValueChanged(ByVal e As ValueChangedEventArgs)
            RaiseEvent ValueChanged(Me, e)
        End Sub

        Private Sub m_owner_PropertyChanged(ByVal sender As Object, ByVal e As PropertyChangedEventArgs)
            If m_dependentValuesToListenFor.ContainsKey(e.PropertyName) Then
                RefreshProperty()
            End If
        End Sub

        Private Sub TriggerValueChanged(ByVal sender As Object, ByVal e As EventArgs)
            RefreshProperty()
        End Sub

        Private Shared ReadOnly s_emptyValueChangedEventArgs As New ValueChangedEventArgs()
        ''' <summary>
        ''' Triggers this property to appear changed and causes its its value to be recomputed the next time it's accessed
        ''' </summary>
        Public Sub RefreshProperty()
            'ensure we retrieve the value anew the next time it is requested
            Me.m_value = New Lazy(Of TValue)(m_getValueProperty)

            OnValueChanged(s_emptyValueChangedEventArgs)
            m_owner.RaisePropertyChanged(Name)
        End Sub
    End Class
End Namespace
