﻿Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Diagnostics
Imports System.Linq
Imports System.Text

Namespace SolSoft.DataBinding
    ''' <summary>
    ''' Class encapsulating a property that raises <see cref="INotifyPropertyChanged.PropertyChanged"/> events on its parent when changed
    ''' </summary>
    ''' <typeparam name="T">Type of the property's value</typeparam>
    <DebuggerDisplay("{Value}")>
    Public Class NotifyProperty(Of T)
        Implements IProperty(Of T)
        Private ReadOnly m_name As String
        Private ReadOnly m_owner As IRaisePropertyChanged
        Private ReadOnly m_itemChangedComparer As IEqualityComparer(Of T)

        ''' <summary>
        ''' Creates a notify property with the given owner and name
        ''' </summary>
        ''' <param name="owner">Object that contains this property, on which <see cref="INotifyPropertyChanged.PropertyChanged"/> events will be raised</param>
        ''' <param name="name">Name of the property within its owner</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal name As String)
            If owner Is Nothing Then
                Throw New NotSupportedException("Null owners (e.g. static properties) are not supported")
            End If
            If String.IsNullOrEmpty(name) Then
                Throw New ArgumentException("No property name was specified", ObjectNamingExtensions.GetName(Function() name))
            End If

            Me.m_owner = owner
            Me.m_name = name
            Me.m_itemChangedComparer = EqualityComparer(Of T).Default 'will be overwritten by other constructors if necessary
        End Sub

        ''' <summary>
        ''' Creates a notify property with the given owner and name
        ''' </summary>
        ''' <param name="owner">Object that contains this property, on which <see cref="INotifyPropertyChanged.PropertyChanged"/> events will be raised</param>
        ''' <param name="name">Name of the property within its owner</param>
        ''' <param name="itemChangedComparer">Comparer that will be used to determine if the value has changed</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal name As String, ByVal itemChangedComparer As IEqualityComparer(Of T))
            Me.New(owner, name, Nothing, itemChangedComparer)
        End Sub

        ''' <summary>
        ''' Creates a notify property with the given owner and name
        ''' </summary>
        ''' <param name="owner">Object that contains this property, on which <see cref="INotifyPropertyChanged.PropertyChanged"/> events will be raised</param>
        ''' <param name="name">Name of the property within its owner</param>
        ''' <param name="initialValue">Initial value of the property</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal name As String, ByVal initialValue As T)
            Me.New(owner, name, initialValue, EqualityComparer(Of T).Default)
        End Sub

        ''' <summary>
        ''' Creates a notify property with the given owner and name
        ''' </summary>
        ''' <param name="owner">Object that contains this property, on which <see cref="INotifyPropertyChanged.PropertyChanged"/> events will be raised</param>
        ''' <param name="name">Name of the property within its owner</param>
        ''' <param name="initialValue">Initial value of the property</param>
        ''' <param name="itemChangedComparer">Comparer that will be used to determine if the value has changed</param>
        Public Sub New(ByVal owner As IRaisePropertyChanged, ByVal name As String, ByVal initialValue As T, ByVal itemChangedComparer As IEqualityComparer(Of T))
            Me.New(owner, name)
            If itemChangedComparer Is Nothing Then
                Throw New ArgumentNullException(ObjectNamingExtensions.GetName(Function() itemChangedComparer))
            End If
            Me.m_itemChangedComparer = itemChangedComparer

            Me.m_value = initialValue
        End Sub

        ''' <summary>
        ''' Gets the name of this property
        ''' </summary>
        Public ReadOnly Property Name() As String Implements IProperty(Of T).Name
            Get
                Return m_name
            End Get
        End Property

        Private m_value As T
        ''' <summary>
        ''' Gets this property's current value
        ''' </summary>
        Public ReadOnly Property Value() As T Implements IProperty(Of T).Value
            Get
                Return m_value
            End Get
        End Property

        ''' <summary>
        ''' Event fired when the property value has changed
        ''' </summary>
        Public Event ValueChanged As EventHandler(Of ValueChangedEventArgs) Implements SolSoft.DataBinding.INotifyValueChanged.ValueChanged

        ''' <summary>
        ''' Raises the <see cref="ValueChanged"/> event
        ''' </summary>
        ''' <param name="e"></param>
        Protected Overridable Sub OnValueChanged(ByVal e As ValueChangedEventArgs)
            Dim v As EventHandler(Of ValueChangedEventArgs) = ValueChangedEvent
            RaiseEvent ValueChanged(Me, e)
            RaiseEvent ValueChanged(Me, e)
        End Sub

        ''' <summary>
        ''' Sets a new value for this property.  If it is different from the current property, the appropriate change events will be raised.
        ''' </summary>
        ''' <param name="newValue">New value to be set</param>
        ''' <returns>True if the value was changed</returns>
        Public Function SetValue(ByVal newValue As T) As Boolean
            Return SetValue(newValue, False)
        End Function

        ''' <summary>
        ''' Sets a new value for this property; if the new value differs from the old value, a PropertyChanged event will be fired from its owner
        ''' </summary>
        ''' <param name="newValue">New value to be set</param>
        ''' <param name="overrideEqualityCheck">Whether or not to override the equality check.  If true, will raise the change events even if the value is unchanged</param>
        ''' <returns>True if the value was changed</returns>
        Public Overridable Function SetValue(ByVal newValue As T, ByVal overrideEqualityCheck As Boolean) As Boolean
            If (Not overrideEqualityCheck) AndAlso CType(m_itemChangedComparer, Object).Equals(m_value, newValue) Then
                Return False
            End If

            m_value = newValue

            OnValueChanged(s_emptyValueChangedArgs)
            m_owner.RaisePropertyChanged(Name)

            Return True
        End Function

        Private Shared ReadOnly s_emptyValueChangedArgs As New ValueChangedEventArgs()
    End Class

End Namespace
